/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
    This file is part of Atmegatron.

    Atmegatron is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Atmegatron is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/

//***15 Biquad filter algorithms***



//These constants are used to define the MAXIMUM amount the env and lfo can effect the filter cutoff.
//The default max (for both env and lfo) is: 
//********multiplier = 4***********
#define FILT_MAX 1.386294361119891f  //ln(multiplier)
#define FILT_MULT 64                 //multipler used for calculating lookup table  (256 / multiplier)  
#define FILT_BS 6UL                  //bitshift amount when multiplying fc by filt lfo amt.  log2(FILT_MULT)

#define SQRT2 1.414213562373095f    //squareroot 2 - useful!

#define MINQ 0.5f                  //minimum value of Q (default = 0.5)
#define MULTQ 0.076470588235294f   //multiplier to convert from filt_q(0-255) to real Q (0.5-20) = (20 - 0.5)/255;

//lets and gets
byte filt_fc = 255;              //filter cutoff - not the actual Fc, just a way to store knob position as byte (0-255)               
byte filt_q = 0;                 //filter resonance - not the actual Q, just a way to store knob position as byte (0-255)
byte filt_fenvamt = 0;           //filter env amount
byte filt_lfoamt = 0;            //filter lfo amount

//local vars
byte filt_lfolookup[256];        //look up table to convert lfo output to Fc multiplier.  Lookup saves having to do v slow exp calculation every cycle
byte filt_fenvlookup[256];       //look up table to convert env output to Fc multiplier.  Lookup saves having to do v slow exp calculation every cycle
float filt_fenvamtf = 0;         //filter env amount as float (0-1)
float filt_lfoamtf = 0;          //filter lfo amount as float (0-1)
float filt_fc_calc = 0;          //final Fc for use in biquad equation (after mult with env, lfo etc)
float filt_q_calc = 1;           //final Q for use in biquad equation (after mult with env, lfo etc)
float two_pi_over_sf= 0;         //2pi / sample freq (ie interrupt freq)
float pi_over_sf= 0;             //pi / sample freq 

// filter constants - these are the constants used by biquad equation.  Whenever Fc, Q or filter type changes, they need recalculating
float a0[2]={1,1};
float a1[2];
float a2[2];
float b0[2];
float b1[2];
float b2[2];

// lets and gets
//Filter cutoff frequency knob position (0-255)
void Filt_Let_Fc(byte newfc){
    filt_fc = newfc;
}
byte Filt_Get_Fc(){
  return filt_fc;
}
//Filter resonance (Q) knob position (0-255)
void Filt_Let_Q(byte newq){
  filt_q = newq;
}
byte Filt_Get_Q(){
  return filt_q;
}


// Filter meat
void Filt_CalcVals(byte oscNum){
	unsigned long max_fc;                          //maximum cutoff frequency.  Must be < Sf/2.  Will ring at Sf/2.
	unsigned long f;                               //intermediate value for fict_fc_calc (before converting to float)


	max_fc = Master_Get_SampleFreq(oscNum) >> 1UL;     //calculate max cutoff frequency.  Sf / 2
	max_fc -= Master_Get_SampleFreq(oscNum) >> 6UL;    //minus a bit more for safety.  reduce maxfc by proportion of Fs (about 1.5%)
    
	f = max_fc;                                  //start with Fc at maximum
	f = f * (filt_fc + 1) >> 8UL;                //then scale by filt cutoff knob position (use >>8 so knob=255 = max_fc)
  
	if (filt_lfoamt>0){                          //if filt lfo knob > 0
		f = f * (Filt_Get_LFOGain() + 1) >> FILT_BS;   //mult fc by lfo gain and bit shift.  This is the fixed point maths way of mult fc by lfo (i.e. not using floating point)
	}

	if (filt_fenvamt>0){
		f = f * (Filt_Get_FenvGain() + 1) >> FILT_BS;   //mult fc by env gain and bit shift.  This is the fixed point maths way of mult fc by env (i.e. not using floating point)
	}
	if (f > max_fc){                              //sort out if f > max_fc
		f = max_fc;         
	}
    
	filt_fc_calc = (float)f;                      //convert to float ready for biquad calculation
	filt_q_calc = (float)filt_q * MULTQ + MINQ;   //calculate q based on knob position. 0.5-20
	pi_over_sf = PI/Master_Get_SampleFreq(oscNum);      //used in biquad equations
	two_pi_over_sf = 2 * pi_over_sf;              //used in biquad equations
    
	LPValCalculator(oscNum);  
    
	//This divide would normally be done in the Biquad_Process. However doing the time consuming divide here means that there's only 5 divides, rather than 32 (in Biquad_Process)
    
	b0[oscNum] /= a0[oscNum];
	b1[oscNum] /= a0[oscNum];
	b2[oscNum] /= a0[oscNum];
	a1[oscNum] /= a0[oscNum];
	a2[oscNum] /= a0[oscNum];


}

// Calculate biquad filter constants for filter
void LPValCalculator(byte oscNum){
    float w = two_pi_over_sf * filt_fc_calc;
    float alpha = sin(w) / filt_q_calc * 0.5;
    float cs = cos(w);
    b0[oscNum] = (1 - cs) / 2;
    b1[oscNum] = 1 - cs;
    b2[oscNum] = b0[oscNum];
    a0[oscNum] = 1 + alpha;
    a1[oscNum] = -2 * cs;
    a2[oscNum] = 1 - alpha;
}

//Process the wavetable
void Filt_Process(byte oscNum){
  
	byte i;
	char out;
	float in, fout, multb, bout;
  

	for (i=0;i<WAVE_LEN;i++){                                          //cycle through each sample
		bout = Biquad_process(oscNum, (float)Wave_Get_Process(i));           //otherwise just get sample and perform biquad process
		if (bout>127){                                                 //constrain biquad output to max/min char value (by clipping)
			out = 127;
		}
		else if (bout<-128){
			out = -128;
		}
		else{
			out = (char)bout;
		}       
		Wave_Let_Process(i,out);                                       //write sample back to wavetable
	}
}
//this is the standard biquad filter equation.  normally would divide by a0, but this has already been done in filt_calcvals, to save the number of time consuming divisions required
float Biquad_process(byte oscNum, float bi0) {
  static float bi1[2];
  static float bi2[2];
  float bo0;
  static float bo1[2];
  static float bo2[2];
    bo0 = b0[oscNum] * bi0 + b1[oscNum] * bi1[oscNum] + b2[oscNum] * bi2[oscNum] - a1[oscNum] * bo1[oscNum] - a2[oscNum] * bo2[oscNum];
    bi2[oscNum] = bi1[oscNum];
    bi1[oscNum] = bi0;
    bo2[oscNum] = bo1[oscNum];
    bo1[oscNum] = bo0;
    return bo0;
}

//Filter LFO amount.  This is stored as a byte representing knob position and a float (0-FILT_MAX)
void Filt_Let_LFOAmt(byte newamt){
  if (newamt!=filt_lfoamt){                                //if new value different to current
    filt_lfoamt = newamt;                                  //set new value
    filt_lfoamtf = (float)filt_lfoamt * FILT_MAX / 255;    //calculate new float value used to calculate LFO lookup table (0-FILT_MAX)
    memset(filt_lfolookup,0,sizeof(filt_lfolookup));       //clear the lookup table  (quicker than loop)
  }
}
byte Filt_Get_LFOAmt(){
  return filt_lfoamt;
}
//Return the 'gain' of LFO.  This is stored in a lookup table to save time consuming calculations each time.  See the forums for further explanation of the 'gain' term
byte Filt_Get_LFOGain(){
  byte index;
  float f, g;
  index = LFO_Get_Level() + 127;                          //get the current output level of LFO (-127 - 128) and add offset (can't lookup negative array indexes)
  if (filt_lfolookup[index]==0){                          //if index of lookup table hasn't yet been calculated:
      f = (float)LFO_Get_Level() / 127;                   //convert LFO output to float (0 - 1)
      g = round(exp(f * filt_lfoamtf) * FILT_MULT - 1);   //calculate lookup table.  e.g. lfoamt = 0: exp(0) * 64 - 1 = 63   e.g.  lfoamt = 1 and lfo level = 127: exp(ln(4)) * 64 - 1 = 255  e.g. lfoamt = 1 and lfo = -127: exp(-ln(4)) * 64 - 1 = 15  
      filt_lfolookup[index] = (byte)g;                    //write value to lookup table
  }
  return filt_lfolookup[index];                           //return lookuptable value
}
//Filter Env amount.   This is stored as a byte representing knob position and a float (0-FILT_MAX)
void Filt_Let_FenvAmt(byte newamt){
  if (newamt!=filt_fenvamt){                                //if new value different to current
    filt_fenvamt = newamt;                                  //set new value
    filt_fenvamtf = (float)filt_fenvamt * FILT_MAX / 255;   //calculate new float value used to calculate env lookup table (0-FILT_MAX)
    memset(filt_fenvlookup,0,sizeof(filt_fenvlookup));      //clear the lookup table  (quicker than loop)
  }
}
byte Filt_Get_FenvAmt(){
  return filt_fenvamt;
}
//Return the 'gain' of env.  This is stored in a lookup table to save time consuming calculations each time.  See the forums for further explanation of the 'gain' term
byte Filt_Get_FenvGain(){
  byte index;
  float f, g;
  index = Fenv_Get_Level() + 127;                          //get the current output level of env (-127 - 128) and add offset (can't lookup negative array indexes)
  if (filt_fenvlookup[index]==0){                          //if index of lookup table hasn't yet been calculated:
      f = (float)Fenv_Get_Level() / 127;                   //convert env output to float (0 - 1)
      g = round(exp(f * filt_fenvamtf) * FILT_MULT - 1);   //calculate lookup table.  e.g. envamt = 0: exp(0) * 64 - 1 = 63   e.g.  envamt = 1 and lfo level = 127: exp(ln(4)) * 64 - 1 = 255  e.g. envamt = 1 and env = -127: exp(-ln(4)) * 64 - 1 = 15  
      filt_fenvlookup[index] = (byte)g;                    //write value to lookup table
  }
  return filt_fenvlookup[index];                           //return lookuptable value
}
