/*
Copyright 2013 Paul Soulsby www.soulsbysynths.com
This file is part of Atmegatron.

Atmegatron is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Atmegatron is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Atmegatron.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <SPI.h>              // Used by MCP23S17 library
#include <MCP23S17.h>         // Used to control the MCP23S17 expander ICs

#define ENCODER_USE_INTERRUPTS
#define ENCODER_OPTIMIZE_INTERRUPTS
//#define ENCODER_DO_NOT_USE_INTERRUPTS
#include <Encoder.h>          //class used to read rotary encoder

#ifndef cbi                  //define clear-bit and set-bit for bit operations
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#define DEBOUNCETICKS 50    //debounce time for buttons i.e. lockout time after press
#define LEDFLASHTICKS 50    //speed of red/green button flash
#define LOCKOUT 3000        //time between ctrl stop moving and locking it out
#define HOLDLENGTH 2000     //time required to hold function button for hold functions (e.g. patch save)

//PROG MEM
prog_uint16_t bg_select_store[16] PROGMEM = {1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768};              //values used to light single led of led array
prog_uint16_t bg_fill_store[16] PROGMEM = {1,3,7,15,31,63,127,255,511,1023,2047,4095,8191,16383,32767,65535};            //values used to fill leds between 0-x of led array
prog_uint16_t envADR_store[16] PROGMEM = {0,8,23,48,88,151,245,381,574,845,1217,1722,2399,3297,4476,6013};               //times used for attack, decay and release of envelopes
prog_uchar envS_store[16] PROGMEM = {0,17,34,51,68,85,102,119,136,153,170,187,204,221,238,255};                          //levels used for sustain of envelope1 (unipolar for amplitude)
prog_char fenvS_store[16] PROGMEM = {0,8,17,25,34,42,51,59,68,76,85,93,102,110,119,127};                                 //levels used for sustain of envelope2 (bipolar for filter)
prog_uint16_t speed_store[16] PROGMEM =  {32767,16383,8191,6143,5460,4095,3071,2730,2047,1535,1023,767,511,383,255,127}; //values used for speeds of LFO and arpeggiator
prog_uint16_t porta_store[16] PROGMEM = {0,5,14,27,46,77,123,195,306,477,740,1145,1768,2728,4207,6484};                  //values for portamento speeds

//hardware
MCP hard_bg_func(0);          //object to control MCP expander for function dial LED array
MCP hard_bg_val(1);           //object to control MCP expander for value dial LED array
MCP hard_bg_led(2);           //object to control MCP expander for button and encoder LEDs
Encoder hard_enc_func(5, 4);  //object to read function rotary encoder (on pins 5&4 - no interrupt, so less responsive)
Encoder hard_enc_val(2, 7);   //object to read value rotary encoder (on pins 2&7 - interrupts, so more responsive)

//lets and gets (not all letable or getable at pres)
byte hard_func = 255;         //current selected function. 255 to allow reset in Init_Hardware
byte hard_val[16];            //current select value for each function.
byte hard_ctrl[2][6];         //current value for each digital control [bank][control]
boolean hard_shift[16];       //current value of shift for each function (red/green mode)

//local variables
byte hard_curctrlbank = 0;        //current value of ctrl bank (red/green mode of controls)
byte hard_curctrl = 1;            //current control being read.  Only 1 control is read in each pass.  read order:  0,1,0,2,0,3,0,4 etc, so cutoff is smooth as poss
byte hard_nextctrl = 1;           //sets next control to be read after cutoff.
byte lastfuncbut = B00000000;     //last value of function button input
byte lastvalbut = B00000000;      //last value of value button input
byte lastbank = B00000000;        //last value of ctrl bank button input
byte lastctrl[6];                 //last value read of each ctrl
long lastval;                     //last value of value rotary encoder
byte maxval = 15;                 //maximum value that is selectable on the value dial (normally 15 apart from user wave mode)
boolean testbutdown = false;      //test button (value encoder button)
boolean movingctrl[6];            //is a ctrl moving (i.e. is user turning it)
unsigned long stopmovingtick[6];  //tick when ctrl will lock out (to stop jitter) assuming user has stopped moving it
unsigned long butstarttick = 0;   //tick when button was pressed
unsigned long debouncetick = 0;   //tick when ctrl/button unlocked from debounce
boolean finishedhold = false;     //used by function button for function that require 2 sec hold (e.g. patch save)
unsigned long ledflashnexttick[2];//tick when red/green led toggle during flash    [0 = function encoder LED, 1 = MIDI LED]
boolean ledflashstate[2];         //current state of flashing LED
byte ledflashtimes[2];            //total number of times LED should flash
byte ledflashcur[2];              //current flash number
boolean ledflashing[2];           //is LED flashing
boolean MIDIchannelmode = false;

//Hardware setup, only run once
void Init_Hardware(){
	byte i,j;
	long newfunc, newval;
	byte newctrl;
	int ctrlread;
	
	//set digital inputs for encoder buttons and ctrl bank button
	DDRB = DDRB & B11111100;        //set digital 8-9 as inputs, leave rest
	DDRD = DDRD & B10111111;        //set digital 6 as input, leave rest
	
	hard_bg_func.pinMode(0x0000);   //set all pins on MCP expander as outputs
	hard_bg_val.pinMode(0x0000);    //set all pins on MCP expander as outputs
	hard_bg_led.pinMode(0x0000);    //set all pins on MCP expander as outputs
	
	Hardware_Let_Function(0);       //initialise function dial to 0
	
	Hardware_LED_SetState(3,HIGH);  //initialise value encoder LED to red
	Hardware_LED_SetState(2,LOW);

	Hardware_Refresh_ValBG();       //refresh value dial LED array

	Hardware_Let_CtrlBank(0);       //initialise ctrl bank to 0 (red mode)
	
	//READ FUNCTION ENCODER BUTTON to see if setting MIDI channel number
	newctrl = ~PINB;                                                //read input pins (quicker than digital read)
	newctrl = newctrl & B00000001;                               //just interested in pin 8
	if (newctrl > B00000000) {                              //if func button down , then MIDI channel mode
		MIDIchannelmode = true;
		Hardware_LED_SetState(3,HIGH);  //initialise value encoder LED to yellow (woo)
		Hardware_LED_SetState(2,HIGH);
		Hardware_Let_Value(FUNC_MEM,Memory_Channel_Read());
	}
	
	for (i=0;i<=5;i++){
		ctrlread = analogRead(i);     //initialise lastctrl array.  use analogread, coz speed not important
		newctrl = ctrlread >> 2;      //all ctrls are stored as byte (0-255) not ints
		lastctrl[i] = newctrl;
	}
	hard_curctrl = 0;
	//analog read is normally done by setting the A/D conversion going, then seeing if it is ready on next pass.  means that CPU isn't stalled waiting for A/D to finish
	ADMUX = (1 << 6) | (hard_curctrl & 0x07);  // set the analog reference (high two bits of ADMUX) and select the channel (low 4 bits).  this also sets ADLAR (left-adjust result) to 0 (the default).
	sbi(ADCSRA, ADSC);                         // start the first conversion
}


//*************POLL HARDWARE***********************
void Hardware_Poll(){
	long newfunc, newval;
	byte r, newbut, newfuncbut, i, newvalbut;
	int newctrl;
	byte  diff;
	unsigned long butcurtick;
	uint8_t low, high;
	
	//  READ FUNCTION ENCODER   -  doesn't use interrupt so uses debounce to try and stop extraneous pulses
	newfunc = hard_enc_func.read();                    //read encoder
	if (master_tick > debouncetick && newfunc!=0) {    //is debounce unlocked
		if (newfunc > 2){                                //has encoder turned more than 2 pulses
			if (hard_func==15){                            //increment function
				Hardware_Let_Function(0);
			}
			else{
				Hardware_Let_Function(hard_func+1);
			}
			hard_enc_func.write(0);                        //set encoder pulse number back to 0
			debouncetick = master_tick + DEBOUNCETICKS;    //set debounce (lock encoder out)
		}
		else if (newfunc < -2) {                         //same as above but for anticlockwise
			if (hard_func==0){
				Hardware_Let_Function(15);
			}
			else{
				Hardware_Let_Function(hard_func-1);
			}
			hard_enc_func.write(0);
			debouncetick = master_tick + DEBOUNCETICKS;
		}
	}
	else {
		hard_enc_func.write(0);                          //if debouncing, just set encoder pulses to 0
	}


	//   READ VALUE ENCODER   -   uses interrupts, so debounce not necessary
	newval = hard_enc_val.read();
	if (newval!=0){                                                            //has encoder moved at all
		if (testbutdown==true){                                                  //if value encoder button is down, change test note pitch
			if (newval > lastval){
				MIDI_TestButtonInc();
			}
			if (newval < lastval){
				MIDI_TestButtonDec();
			}
			lastval = newval;
		}
		else{
			if (newval >= 4){                                                      //has encoder moved 4 pulses (= 1 click)
				if (hard_val[hard_func]==maxval && IsFuncFill(hard_func)==false){    //if not a fill function (fill functions are A,D,S,R & portamento) and value at max
					Hardware_Let_Value(hard_func, 0);                                  //go back to 0
				}
				else if (hard_val[hard_func]<maxval){                                //otherwise increment value
					Hardware_Let_Value(hard_func, hard_val[hard_func]+1);
				}
				hard_enc_val.write(0);                                               //set encoder pulses back to 0
			}
			else if (newval <= -4) {                                               //same as above but for anti-clockwise
				if (hard_val[hard_func]==0 && IsFuncFill(hard_func)==false){
					Hardware_Let_Value(hard_func, maxval);
				}
				else if (hard_val[hard_func]>0){
					Hardware_Let_Value(hard_func, hard_val[hard_func]-1);
				}
				hard_enc_val.write(0);
			}
		}
	}


	//READ CTRL BANK BUTTON  (red/green mode)
	newbut = ~PIND;                                        //read input pins (quicker than digital read)
	newbut = newbut & B01000000;                           //just interested in pin 6
	if (master_tick > debouncetick) {                      //if not debouncing (button locked out)
		if (lastbank!=newbut && newbut>B00000000){         //has a button just been pressed - do nothing on release
			if (hard_curctrlbank==0){                        //toggle ctrl bank
				Hardware_Let_CtrlBank(1);
			}
			else{
				Hardware_Let_CtrlBank(0);
			}
			for (i=0;i<6;i++){                                //set all ctrls to not moving, just in case last thing you did was move ctrl in other bank and not locked out
				movingctrl[i] = false;
				stopmovingtick[i] = master_tick + LOCKOUT;
			}
			debouncetick = master_tick + DEBOUNCETICKS;        //reset debounce too
		}
	}
	lastbank = newbut;                                       //store last state of button
	
	
	//READ FUNCTION ENCODER BUTTON
	newbut = ~PINB;                                                //read input pins (quicker than digital read)
	newfuncbut = newbut & B00000001;                               //just interested in pin 8
	if (master_tick > debouncetick) {                              //if not debouncing (button locked out)
		if (IsShiftHold(hard_func)==true){                           //and we on a function that requires function button to be held (e.g. save patch)
			//********SAVE AND LOAD IS DONE HERE**********
			if (lastfuncbut==B00000000 && newfuncbut>B00000000){       //button pressed down, go green and wait
				butstarttick = master_tick;
				Hard_Let_Shift(hard_func, true);
				finishedhold = false;
				debouncetick = master_tick + DEBOUNCETICKS;              //reset debounce
			}
			else if (lastfuncbut>B00000000 && newfuncbut>B00000000){   //is button still held
				butcurtick = master_tick - butstarttick;
				if (butcurtick >= HOLDLENGTH && finishedhold == false){  //has button been held down for longer than hold length
					switch (hard_func){
						case FUNC_WAVEA:
						Memory_UserWave_Write(hard_val[FUNC_WAVEA]);        // then do userwave write
						break;
						case FUNC_AENVA:
						break;
						case FUNC_MEM:
						if (MIDIchannelmode==true){
							MIDI_Set_Channel(hard_val[hard_func]);
						}
						else{
							Memory_Save(hard_val[hard_func]);                  // then save patch
						}
						break;
					}
					finishedhold = true;                                   // completed operation (to prevent refires)
					Hard_Let_Shift(hard_func, false);                      // make encoder flash
					Hardware_LED_StartFlash(0,7);
				}
				debouncetick = master_tick + DEBOUNCETICKS;
			}
			else if (newfuncbut==B00000000 && Hard_Get_Shift(hard_func)==true){  //has button been released
				butcurtick = master_tick - butstarttick;
				if (butcurtick < HOLDLENGTH){                                      //and it's been held less time than hold length
					switch (hard_func){
						case FUNC_WAVEA:
						Memory_UserWave_Read(hard_val[FUNC_WAVEA]);                   //then toggle user wave mode
						break;
						case FUNC_MEM:
						Memory_Load(hard_val[hard_func]);                            //then load patch
						for(i=0;i<6;i++){
							movingctrl[i] = false;                                     //stop knob jitter from overriding loaded patch
						}
						break;
					}
				}
				Hard_Let_Shift(hard_func, false);                                  //set shift back to false, LED red (red mode)
				debouncetick = master_tick + DEBOUNCETICKS;
			}
		}
		else  {
			if (lastfuncbut!=newfuncbut && newfuncbut>B00000000){                //button pressed, but not hold function
				Hard_Let_Shift(hard_func, !Hard_Get_Shift(hard_func));             //then toggle shift  (red/green mode)
				debouncetick = master_tick + DEBOUNCETICKS;                        //reset debounce
			}
		}
	}
	lastfuncbut = newfuncbut;                                                //store last button state
	
	//READ VALUE ENCODER BUTTON
	newvalbut = newbut & B00000010;                         //just interested in pin 6
	if (lastvalbut==B00000000 && newvalbut>B00000000){      //has button been pressed down
		MIDI_TestButtonDown();                                //midi test note on
		testbutdown = true;
		hard_enc_val.write(0);                                //reset encoder pulse, because different ranges for midi test note mode.
		Hardware_LED_SetState(2,HIGH);                        //knob goes green
		Hardware_LED_SetState(3,LOW);
	}
	if (newvalbut==B00000000 && lastvalbut>B00000000){      //has button been released
		MIDI_TestButtonUp();                                  //midi test note off
		testbutdown = false;
		hard_enc_val.write(0);                                //reset encoder pulse again, to stop possible value dial jump
		Hardware_LED_SetState(3,HIGH);                        //knob back to red
		Hardware_LED_SetState(2,LOW);
	}
	lastvalbut = newvalbut;                                 //store last button state
	
	// READ CTRL KNOBS (ONLY READ ONE CTRL ON EACH PASS)
	if (master_tick >= stopmovingtick[hard_curctrl]){       //firstly has enough time passed to lock ctrl (to prevent jitter in DA readings effecting value)
		movingctrl[hard_curctrl] = false;
	}
	if (bit_is_set(ADCSRA, ADSC)==false){                   //is reading from DA is ready
		low  = ADCL;
		high = ADCH;
		newctrl = ((high << 8) | low) >> 2;                   //new reading of ctrl
		if (newctrl > lastctrl[hard_curctrl]){                //calculate difference between new reading and ol
			diff = newctrl-lastctrl[hard_curctrl];
		}
		else{
			diff = lastctrl[hard_curctrl]-newctrl;
		}
		if (diff > 4){                                        //if difference > x then ctrl is being moved.  ctrl lockout sensitivity set here (default = 4)
			movingctrl[hard_curctrl] = true;
			stopmovingtick[hard_curctrl] = master_tick + LOCKOUT;            //ctrl is moving, so reset lockout time
		}
		if (movingctrl[hard_curctrl] == true){
			Hardware_Let_Ctrl(hard_curctrlbank, hard_curctrl, newctrl);      //pass new value to ctrl parameter
			lastctrl[hard_curctrl] = newctrl;                                //store last ctrl state
		}
		
		if(hard_curctrl==0){                                  //increment current ctrl to read.  order is: 0,1,0,2,0,3,0,4.. so filter cutoff ctrl is smooth
			hard_curctrl = hard_nextctrl;
			hard_nextctrl++;
			if (hard_nextctrl>5){
				hard_nextctrl = 1;
			}
		}
		else{
			hard_curctrl = 0;
		}
		ADMUX = (1 << 6) | (hard_curctrl & 0x07);             //start DA conversion of next ctrl going
		sbi(ADCSRA, ADSC);
	}
	
	Hardware_LED_RefreshFlash();                            //update LEDs that may be flashing

}



//**************CONTROL KNOBS*************************
//Let ctrl parameter equal new value
void Hardware_Let_Ctrl(byte shift, byte index, byte value){
	switch (index){
		case CTRL_FILT:                      //Filter Cutoff frequency (same in red and green mode)
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Filt_Let_Fc(value);
			}
			else{
				Pitch_Let_Offset(0,value);
			}
			break;
		case CTRL_Q:                        //Filter resonance (same in red and green mode)
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Filt_Let_Q(value);
			}
			else{
				Pitch_Let_Offset(1,value);
			}
			break;
		case CTRL_ENV:
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Filt_Let_FenvAmt(value);        //Filter envelope amount
			}
			else {
				Pitch_Let_FenvAmt(value);       //Pitch envelope amount
			}
			break;
		case CTRL_LFO:
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Filt_Let_LFOAmt(value);         //Filter LFO amount
			}
			else {
				Pitch_Let_LFOAmt(value);        //Pitch LFO amount
			}
			break;
		case CTRL_AMP:
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Amplitude_Let_LFOAmt(value);    //Amplitude LFO amount
			}
			else {
				PWM_Let_LFOAmt(value>>3);       //Pulse width modulation amount
			}
			break;
		case CTRL_FX:
			hard_ctrl[shift][index] = value;
			if (shift==0){
				Distortion_Let_Amt(value>>5);  //Distortion (takes values 0-7)
			}
			else {
				Wave_Let_OscMix(0,15-(value>>4));
				Wave_Let_OscMix(1,value>>4);
			}
			break;
	}
}
byte Hardware_Get_Ctrl(byte shift, byte index){    //get current ctrl value
	return hard_ctrl[shift][index];
}

//**************FUNCTION KNOB*******************
//set current function
void Hardware_Let_Function(byte newfunc){
	if (newfunc!=hard_func){                                      //if new value is different to current
		hard_func = newfunc;
		Hardware_Refresh_FuncBG();                                  //refresh function dial LED array
		Hardware_Refresh_ValBG();                                   //refresh value dial LED array (to show val of new func)
		Hardware_Refresh_ShiftLED(hard_func);                       //refresh shift (red/green) mode (to show shift mode of new func)
		if (hard_shift[FUNC_AENVS]==true && hard_func==FUNC_WAVEA){  //set max value of value dial (normally 15)
			maxval = 5;
		}
		else{
			maxval = 15;
		}
	}
}

byte Hardware_Get_Function(){                                  //get current function
	return hard_func;
}

void Hardware_Refresh_FuncBG(){
	hard_bg_func.digitalWrite(pgm_read_word_near(&(bg_select_store[hard_func])));  //set function dial LED array (using MCP 0)
}

//*************VALUE KNOB***********************
//set new value of function
void Hardware_Let_Value(byte func, byte newval){
	if (newval!=hard_val[func]){                      //is new value different to current value
		hard_val[func] = newval;
		Hardware_Adjust_Value(func, hard_val[func]);    //the meat of how the new value effects the sound
		if (func==hard_func){
			Hardware_Refresh_ValBG();                     //refresh the value dial LED array
		}
	}
}

byte Hardware_Get_Value(byte func){                //get the current value of a function
	return hard_val[func];
}

void Hardware_Adjust_Value(byte func, byte val){               //the meat of how a new value effects the sound
	switch (func) {
		case FUNC_WAVEA:
		if (hard_shift[FUNC_AENVS]==false){                    //if not user wave mode
			Wave_Let_Table(0, val);                                 //select waveform
		}
		break;
		case FUNC_WAVEB:
		if (hard_shift[FUNC_AENVR]==false){                    //if not user wave mode
			Wave_Let_Table(1, val);                                 //select waveform
		}                                  //set filter type
		break;
		case FUNC_FENVA:
		Fenv_Let_A(pgm_read_word(&(envADR_store[val])));       //set filter/pitch envelope attack time  (stored in PROGMEM)
		break;
		case FUNC_FENVDR:
		Fenv_Let_DR(pgm_read_word(&(envADR_store[val])));      //set filter/pitch envelope decay time
		break;
		case FUNC_FENVS:
		Fenv_Let_S(pgm_read_byte_near(&(fenvS_store[val])));   //set filter/pitch envelope sustain level
		break;
		case FUNC_AENVA:
		Aenv_Let_A(pgm_read_word(&(envADR_store[val])));       //set amplitude envelope attack time
		break;
		case FUNC_AENVD:
		Aenv_Let_D(pgm_read_word(&(envADR_store[val])));       //set amplitude envelope decay time
		break;
		case FUNC_AENVS:
		Aenv_Let_S(pgm_read_byte(&(envS_store[val])));         //set amplitude envelope sustain level
		break;
		case FUNC_AENVR:
		Aenv_Let_R(pgm_read_word(&(envADR_store[val])));       //set amplitude envelope release time
		break;
		case FUNC_LFOTYPE:
		LFO_Let_Type(val);                                     //select lfo waveform
		break;
		case FUNC_LFOSPEED:
		LFO_Let_Speed(pgm_read_word_near(&(speed_store[val])));//select lfo speed  (stored in PROGMEM)
		break;
		case FUNC_ARPTYPE:                                       //select arpeggiator pattern
		Arp_Let_Type(val);
		break;
		case FUNC_ARPSPEED:
		Arp_Let_Speed(pgm_read_word_near(&(speed_store[val])));            //set arpeggiator speed when free running
		MIDI_Let_ClockArpSpeed(pgm_read_word_near(&(speed_store[val])));   //set arpeggiator speed when MIDI clock synced
		break;
		case FUNC_PORTA:
		Pitch_Let_Porta(pgm_read_word_near(&(porta_store[val])));          //set portamento time
		break;
		case FUNC_BITCRUSH:
		BitCrush_Let_Type(val);                                //set wave crusher setting
		break;
		case FUNC_MEM:                                           //this is a hold function, so do nothing (patch save)
		break;
	}
}

void Hardware_Refresh_ValBG(){                                //refresh value dial LED array
	if (IsFuncFill(hard_func)==true){
		hard_bg_val.digitalWrite(pgm_read_word_near(&(bg_fill_store[hard_val[hard_func]])));
	}
	else{
		hard_bg_val.digitalWrite(pgm_read_word_near(&(bg_select_store[hard_val[hard_func]])));
	}
}

//*************CTRL BANK BUTTON***********************
void Hardware_Let_CtrlBank(byte value){        //set ctrl bank (red/green mode)
	hard_curctrlbank = value;
	Hardware_Refresh_CtrlBankLED();              //refresh LED colour
}

//not used at present so remove it
//byte Hardware_Get_CtrlBank(){
//  return hard_curctrlbank;
//}

void Hardware_Refresh_CtrlBankLED(){
	
	if (hard_curctrlbank==1){
		hard_bg_led.digitalWrite(5,LOW);    //LED is green
		hard_bg_led.digitalWrite(6,HIGH);
	}
	else{
		hard_bg_led.digitalWrite(5,HIGH);  //LED is red
		hard_bg_led.digitalWrite(6,LOW);
	}
}

//***********SHIFT BUTTON**************
//let function encoder button equal new value
void Hard_Let_Shift(byte func, boolean newshift){
	if (newshift!=hard_shift[func]){
		switch (func) {
			case FUNC_FENVA:                          //tie filter/pitch envelope shifts together (envelope invert mode)
			case FUNC_FENVDR:
			case FUNC_FENVS:
			hard_shift[FUNC_FENVA] = newshift;
			hard_shift[FUNC_FENVDR] = newshift;
			hard_shift[FUNC_FENVS] = newshift;
			break;
			case FUNC_LFOTYPE:                      //tie LFO shifts together (LFO invert mode)
			case FUNC_LFOSPEED:
			hard_shift[FUNC_LFOTYPE] = newshift;
			hard_shift[FUNC_LFOSPEED] = newshift;
			break;
			case FUNC_ARPTYPE:                      //tie arpeggiator shifts together (arpeggiator ping pong mode)
			case FUNC_ARPSPEED:
			hard_shift[FUNC_ARPTYPE] = newshift;
			hard_shift[FUNC_ARPSPEED] = newshift;
			break;
			default:
			hard_shift[func] = newshift;
		}
		Hardware_Adjust_Shift(func, newshift);      //meat of how shift affects sound
		if (func==hard_func){
			Hardware_Refresh_ShiftLED(func);          //refresh function knob LED colour
		}
	}
}
boolean Hard_Get_Shift(byte func){              //get shift value of a function
	return hard_shift[func];
}
void Hardware_Refresh_ShiftLED(byte func){      //refresh colour of function knob LED
	
	if (hard_shift[func]==true){
		hard_bg_led.digitalWrite(1,LOW);            //green
		hard_bg_led.digitalWrite(0,HIGH);
	}
	else{
		hard_bg_led.digitalWrite(1,HIGH);          //red
		hard_bg_led.digitalWrite(0,LOW);
	}
}
void Hardware_Adjust_Shift(byte func, boolean val){    //meat of how shift affects sound
	byte i;
	switch (func) {
		case FUNC_WAVEA:
		if (hard_shift[FUNC_AENVS]==false){
			if (val==true){
				Wave_Let_Bank(0,1);                          //set waveform bank
			}
			else{
				Wave_Let_Bank(0,0);
			}
		}
		break;
		case FUNC_WAVEB:
		//Filt_Let_GainAdj(val);                          //set filter normalise mode
		if (hard_shift[FUNC_AENVR]==false){
			if (val==true){
				Wave_Let_Bank(1,1);                          //set waveform bank
			}
			else{
				Wave_Let_Bank(1,0);
			}
		}
		break;
		case FUNC_FENVA:
		case FUNC_FENVDR:
		case FUNC_FENVS:
		Fenv_Let_Invert(!Fenv_Get_Invert());            //set filter/pitch envelope invert mode
		break;
		case FUNC_AENVA:                                  //hold function, so do nothing here  (SEE Hardware_Poll)
		break;
		case FUNC_AENVD:
		break;
		case FUNC_AENVS:
		Wave_Let_UserMode(0,val);                         //set user wave mode
		if (val==true && hard_val[FUNC_WAVEA]>5){
			Hardware_Let_Value(FUNC_WAVEA,5);
		}
		break;
		case FUNC_AENVR:                                  //hold function, so do nothing here  (SEE Hardware_Poll)
		Wave_Let_UserMode(1,val);                         //set user wave mode
		if (val==true && hard_val[FUNC_WAVEB]>5){
			Hardware_Let_Value(FUNC_WAVEB,5);
		}
		break;
		case FUNC_LFOTYPE:
		case FUNC_LFOSPEED:
		LFO_Let_Invert(val);                            //set LFO invert mode
		break;
		case FUNC_ARPTYPE:
		case FUNC_ARPSPEED:
		Arp_Let_PingPong(val);                          //set arpeggiator ping pong mode
		break;
		case FUNC_PORTA:
		Pitch_Let_PropPorta(val);                       //set proportional portamento mode
		break;
		case FUNC_BITCRUSH:
		BitCrush_Let_PreFilt(val);                      //set wave crusher pre-filter mode
		case FUNC_MEM:                                    //hold function, so do nothing here  (SEE Hardware_Poll)
		break;
	}
}

//Flashing LEDS
void Hardware_LED_SetState(byte LEDnum, byte LEDstate){
	//0 = Function encoder green
	//1 = Function encoder red
	//2 = Value encoder green
	//3 = Value encoder red
	//4 = MIDI
	//5 = Ctrl bank red  // I think turning red and green on together will result in both off, because both legs at 5V.
	//6 = Ctrl bank green
	hard_bg_led.digitalWrite(LEDnum,LEDstate);
}
void Hardware_LED_StartFlash(byte LEDnum, byte FlashTimes){  //flash led
	//0 = Function encoder
	//1 = MIDI
	if (LEDnum==0){                          //set initial state
		hard_bg_led.digitalWrite(1,LOW);
		hard_bg_led.digitalWrite(0,HIGH);
	}
	else{
		hard_bg_led.digitalWrite(4,HIGH);
	}
	ledflashstate[LEDnum] = true;            //value toggles every flash
	ledflashtimes[LEDnum] = FlashTimes;      //set number of times to flash
	ledflashcur[LEDnum] = 0;                 //current flash number = 0
	ledflashing[LEDnum] = true;              //set led is flashing
	ledflashnexttick[LEDnum] = master_tick + LEDFLASHTICKS;    //set next tick for flash to toggle
}
void Hardware_LED_RefreshFlash(){                  //update flashing leds
	byte i;
	for(i=0;i<=1;i++){
		if (ledflashing[i]==true){                    //if flashing
			if (master_tick>=ledflashnexttick[i]){      //and it's time to toggle
				ledflashstate[i] = !ledflashstate[i];     //toggle
				if(ledflashstate[i]==true){               //update state
					if (i==0){
						hard_bg_led.digitalWrite(1,LOW);
						hard_bg_led.digitalWrite(0,HIGH);
					}
					else{
						hard_bg_led.digitalWrite(4,HIGH);
					}
				}
				else{
					if (i==0){
						hard_bg_led.digitalWrite(1,HIGH);
						hard_bg_led.digitalWrite(0,LOW);
					}
					else{
						hard_bg_led.digitalWrite(4,LOW);
					}
				}
				ledflashcur[i]++;                          //increment flash counter
				if(ledflashcur[i]>=ledflashtimes[i]){      //if finished flashing
					ledflashing[i] = false;
					if (i==0){
						Hardware_Refresh_ShiftLED(hard_func);  //go back to the colour you were at
					}
				}
				else{
					ledflashnexttick[i] = master_tick + LEDFLASHTICKS;    //otherwise set next tick for flash toggle
				}
			}
		}
	}
}


boolean IsFuncFill(byte func){
	//returns true for functions where value dial fills up
	switch (func) {
		case FUNC_WAVEA:
		case FUNC_WAVEB:
		case FUNC_LFOTYPE:
		case FUNC_LFOSPEED:
		case FUNC_ARPTYPE:
		case FUNC_ARPSPEED:
		case FUNC_BITCRUSH:
		case FUNC_MEM:
		return false;
		break;
		default:
		return true;
	}
}
boolean IsShiftHold(byte func){
	//returns true for functions where encoder button should be held down
	switch (func) {
		case FUNC_WAVEA:
		return hard_shift[FUNC_AENVS];
		break;
		case FUNC_WAVEB:
		return hard_shift[FUNC_AENVR];
		break;
		case FUNC_AENVA:
		case FUNC_MEM:
		return true;
		break;
		default:
		return false;
	}
}

